// replicated state machine interface.

#ifndef rsm_h
#define rsm_h

#include <string>
#include <vector>
#include "rsm_protocol.h"
#include "rsm_state_transfer.h"
#include "rpc.h"
#include <arpa/inet.h>
#include "paxos.h"



class rsm : public paxos_view_change {
 private:
  void reg1(int proc, handler *);
 protected:
  std::map<int, handler *> procs;
  paxos *pxs;
  class rsm_state_transfer *stf;
  std::map<std::string, rpcc *> cl;
  rpcs *rsmrpc;
  viewstamp myvs;
  viewstamp last_myvs;
  std::string master;
  //if stable is false, this node's rsm state 
  //might not be up-to-date due to view changes
  bool stable; 


  rsm_client_protocol::status client_members(int i, 
               std::vector<std::string> &r);
  rsm_protocol::status invoke(int proc, viewstamp vs,
            std::string mreq, int &dummy);
  rsm_protocol::status transferreq(std::string src, viewstamp vs,
           std::string &r);
  rsm_protocol::status viewstampreq(unsigned int vid, viewstamp &r);

  pthread_mutex_t rsm_mutex;
  pthread_cond_t recovery_cond;

  std::string execute(int procno, std::string req);
  rsm_client_protocol::status client_invoke(int procno, std::string req, 
              std::string &r);
  bool statetransfer(std::string m, viewstamp vs);
  std::string get_master();
  bool find_master(viewstamp &vs);

 public:
  rsm (std::string _first, std::string _me);
  ~rsm() {};

  bool amimaster();
  void set_state_transfer(rsm_state_transfer *_stf) { stf = _stf; };
  void recovery();
  void view_change();

  template<class S, class A1, class R>
    void reg(int proc, S*, int (S::*meth)(const A1 a1, R &));
  template<class S, class A1, class A2, class R>
    void reg(int proc, S*, int (S::*meth)(const A1 a1, const A2 a2, R &));
  template<class S, class A1, class A2, class A3, class R>
    void reg(int proc, S*, int (S::*meth)(const A1 a1, const A2 a2, 
            const A3 a3, R &));
  template<class S, class A1, class A2, class A3, class A4, class R>
    void reg(int proc, S*, int (S::*meth)(const A1 a1, const A2 a2, 
            const A3 a3, const A4 a4, R &));
  template<class S, class A1, class A2, class A3, class A4, class A5, class R>
    void reg(int proc, S*, int (S::*meth)(const A1 a1, const A2 a2, 
            const A3 a3, const A4 a4, 
            const A5 a5, R &));
};

template<class S, class A1, class R> void
  rsm::reg(int proc, S*sob, int (S::*meth)(const A1 a1, R & r))
{
  class h1 : public handler {
  private:
    S * sob;
    int (S::*meth)(const A1 a1, R & r);
  public:
  h1(S *xsob, int (S::*xmeth)(const A1 a1, R & r))
      : sob(xsob), meth(xmeth) { }
    int fn(unmarshall &args, marshall &ret) {
      A1 a1;
      R r;
      args >> a1;
      assert(args.okdone());
      int b = (sob->*meth)(a1,r);
      ret << r;
      return b;
    }
  };
  reg1(proc, new h1(sob, meth));
}

template<class S, class A1, class A2, class R> void
  rsm::reg(int proc, S*sob, int (S::*meth)(const A1 a1, const A2 a2, R & r))
{
 class h1 : public handler {
  private:
    S * sob;
    int (S::*meth)(const A1 a1, const A2 a2, R & r);
  public:
  h1(S *xsob, int (S::*xmeth)(const A1 a1, const A2 a2, R & r))
    : sob(xsob), meth(xmeth) { }
    int fn(unmarshall &args, marshall &ret) {
      A1 a1;
      A2 a2;
      R r;
      args >> a1;
      args >> a2;
      assert(args.okdone());
      int b = (sob->*meth)(a1,a2,r);
      ret << r;
      return b;
    }
  };
  reg1(proc, new h1(sob, meth));
}

template<class S, class A1, class A2, class A3, class R> void
  rsm::reg(int proc, S*sob, int (S::*meth)(const A1 a1, const A2 a2, 
             const A3 a3, R & r))
{
 class h1 : public handler {
  private:
    S * sob;
    int (S::*meth)(const A1 a1, const A2 a2, const A3 a3, R & r);
  public:
  h1(S *xsob, int (S::*xmeth)(const A1 a1, const A2 a2, const A3 a3, R & r))
    : sob(xsob), meth(xmeth) { }
    int fn(unmarshall &args, marshall &ret) {
      A1 a1;
      A2 a2;
      A3 a3;
      R r;
      args >> a1;
      args >> a2;
      args >> a3;
      assert(args.okdone());
      int b = (sob->*meth)(a1,a2,a3,r);
      ret << r;
      return b;
    }
  };
  reg1(proc, new h1(sob, meth));
}

template<class S, class A1, class A2, class A3, class A4, class R> void
  rsm::reg(int proc, S*sob, int (S::*meth)(const A1 a1, const A2 a2, 
             const A3 a3, const A4 a4, R & r))
{
 class h1 : public handler {
  private:
    S * sob;
    int (S::*meth)(const A1 a1, const A2 a2, const A3 a3, const A4 a4, R & r);
  public:
  h1(S *xsob, int (S::*xmeth)(const A1 a1, const A2 a2, const A3 a3, 
            const A4 a4, R & r))
    : sob(xsob), meth(xmeth) { }
    int fn(unmarshall &args, marshall &ret) {
      A1 a1;
      A2 a2;
      A3 a3;
      A4 a4;
      R r;
      args >> a1;
      args >> a2;
      args >> a3;
      args >> a4;
      assert(args.okdone());
      int b = (sob->*meth)(a1,a2,a3,a4,r);
      ret << r;
      return b;
    }
  };
  reg1(proc, new h1(sob, meth));
}


template<class S, class A1, class A2, class A3, class A4, class A5, class R> void
  rsm::reg(int proc, S*sob, int (S::*meth)(const A1 a1, const A2 a2, 
             const A3 a3, const A4 a4, 
             const A5 a5, R & r))
{
 class h1 : public handler {
  private:
    S * sob;
    int (S::*meth)(const A1 a1, const A2 a2, const A3 a3, const A4 a4, 
       const A5 a5, R & r);
  public:
  h1(S *xsob, int (S::*xmeth)(const A1 a1, const A2 a2, const A3 a3, 
            const A4 a4, const A5 a5, R & r))
    : sob(xsob), meth(xmeth) { }
    int fn(unmarshall &args, marshall &ret) {
      A1 a1;
      A2 a2;
      A3 a3;
      A4 a4;
      A5 a5;
      R r;
      args >> a1;
      args >> a2;
      args >> a3;
      args >> a4;
      assert(args.okdone());
      int b = (sob->*meth)(a1,a2,a3,a4,a5,r);
      ret << r;
      return b;
    }
  };
  reg1(proc, new h1(sob, meth));
}

#endif /* rsm_h */
