#include "rsm_client.h"
#include <vector>
#include <arpa/inet.h>



rsm_client::rsm_client(std::string dst)
{
  printf("create rsm_client\n");
  std::vector<std::string> mems;

  sockaddr_in dstsock;
  make_sockaddr(dst.c_str(), &dstsock);
  master.id = dst;
  master.cl = new rpcc(dstsock);
  int ret = master.cl->bind(rpcc::to(1000));
  if (ret < 0) {
    printf("rsm_client::rsm_client bind failure %d failure w %s; exit\n", ret, 
     master.id.c_str());
    exit(1);
  }
  assert (init_members(true));
}

void
rsm_client::master_failure()
{
}

rsm_protocol::status
rsm_client::invoke(int proc, std::string req, std::string &rep)
{
  int ret;
  while (1) {
    printf("rsm_client::invoke proc %x master %s\n", proc, master.id.c_str());
    ret = master.cl->call(rsm_client_protocol::invoke, proc, req, 
        rep, rpcc::to(5000));
    printf("rsm_client::invoke proc %x master %s done %d\n", proc, 
     master.id.c_str(), ret);
    if (ret == rsm_client_protocol::OK) {
      break;
    }
    // you have to handle the cases when the RSM replies BUSY, NOTMASTER or simply fails.
  }
  return ret;
}

bool
rsm_client::init_members(bool send_member_rpc)
{
  if (send_member_rpc) {
    int ret = master.cl->call(rsm_client_protocol::members, 0, known_mems, 
            rpcc::to(1000)); 
    if (ret != rsm_protocol::OK)
      return false;
  }
  if (known_mems.size() < 1) {
    printf("rsm_client::init_members do not know any members!\n");
    assert(0);
  }

  std::string new_master = known_mems.back();
  known_mems.pop_back();

  if (new_master != master.id) {
    sockaddr_in dstsock;
    make_sockaddr(new_master.c_str(), &dstsock);
    master.id = new_master;
    if (master.cl) {
      delete master.cl; 
    }
    master.cl = new rpcc(dstsock);

    if (master.cl->bind(rpcc::to(1000)) < 0) {
      printf("rsm_client::rsm_client cannot bind to master\n");
      return false;
    }
  }
  return true;
}

